"""Low-level Modbus logic."""

import asyncio
import logging
from typing import TypeVar

import tenacity
from tenacity import AsyncRetrying, retry_if_exception_type, stop_after_attempt, stop_after_delay, wait_exponential
from tmodbus import AsyncModbusClient, AsyncRtuTransport, AsyncSmartTransport, AsyncTcpTransport
from tmodbus.exceptions import ModbusResponseError, TModbusError
from tmodbus.utils.crc import calculate_crc16

from .exceptions import ReadException
from .modbus_pdu import (
    CompleteUploadPDU,
    LoginPDU,
    LoginRequestChallengePDU,
    StartFileUploadPDU,
    UploadFileFramePDU,
)
from .register_client import RegisterAwareModbusClient

_LOGGER = logging.getLogger(__name__)

T = TypeVar("T")
RT = TypeVar("RT")


DEFAULT_TCP_PORT = 502
DEFAULT_BAUDRATE = 9600

DEFAULT_UNIT_ID = 0
DEFAULT_TIMEOUT = 10  # especially the SDongle can react quite slowly
DEFAULT_WAIT = 1
DEFAULT_COOLDOWN_TIME = 0.05
WAIT_FOR_CONNECTION_TIMEOUT = 5
WAIT_FOR_LOGIN_TIMEOUT = 5

HEARTBEAT_REGISTER = 49999

FILE_UPLOAD_MAX_RETRIES = 6
FILE_UPLOAD_RETRY_TIMEOUT = 10


RECONNECT_RETRY_STRATEGY = AsyncRetrying(
    wait=wait_exponential(multiplier=1, min=1, max=10),
    # Stop trying to reconnect if the connection has not been re-established within 1 minute
    stop=stop_after_delay(60),
    after=lambda retry_call_state: _LOGGER.debug(
        "Backing off before reconnect for %0.1fs after %d tries",
        retry_call_state.upcoming_sleep,
        retry_call_state.attempt_number,
    ),
)


def log_invalid_response(retry_state: "tenacity.RetryCallState") -> None:
    """Log an invalid response."""
    if retry_state.outcome:
        if e := retry_state.outcome.exception():
            _LOGGER.debug(
                "Backing off for %0.1fs after exception response %s",
                retry_state.upcoming_sleep,
                e,
            )
        else:
            _LOGGER.debug(
                "Backing off for %0.1fs after invalid response %s",
                retry_state.upcoming_sleep,
                retry_state.outcome.result(),
            )
    else:
        _LOGGER.debug(
            "Backing off for %0.1fs before retrying request",
            retry_state.upcoming_sleep,
        )


RESPONSE_RETRY_STRATEGY = AsyncRetrying(
    wait=wait_exponential(multiplier=1, min=1, max=10),
    # Retry up to 3 times on invalid response
    stop=stop_after_attempt(3),
    retry=retry_if_exception_type(TimeoutError),
    reraise=True,
    after=log_invalid_response,
)


class AsyncHuaweiSolarClient(RegisterAwareModbusClient, AsyncModbusClient):
    """Async client to Huawei Solar devices."""

    def for_unit_id(self, unit_id: int) -> "AsyncHuaweiSolarClient":
        """Get a copy of this client for a different unit ID."""
        if unit_id == self.unit_id:
            return self
        return AsyncHuaweiSolarClient(self.transport, unit_id=unit_id, word_order=self.word_order)

    async def get_file(
        self,
        file_type: int,
        customized_data: bytes | None = None,
    ) -> bytes:
        """Read a 'file' via Modbus.

        As defined by the 'Uploading Files' process described in 6.3.7.1 of
        the Solar Inverter Modbus Interface Definitions PDF.
        """
        _LOGGER.debug(
            "Reading file %#x from server %d",
            file_type,
            self.unit_id,
        )
        # Start the upload
        start_upload_response = await self.execute(
            StartFileUploadPDU(
                file_type=file_type,
                customised_data=customized_data or b"",
            ),
        )

        file_length = start_upload_response.file_length
        data_frame_length = start_upload_response.data_frame_length

        # Request the data in 'frames'

        file_data: bytes = b""
        next_frame_no = 0

        while (next_frame_no * data_frame_length) < file_length:
            data_upload_response = await self.execute(
                UploadFileFramePDU(file_type=file_type, frame_no=next_frame_no),
            )

            file_data += data_upload_response.frame_data
            next_frame_no += 1

        # Complete the upload and check the CRC
        file_crc = await self.execute(
            CompleteUploadPDU(file_type=file_type),
        )

        # swap upper and lower two bytes to match how computeCRC works
        swapped_crc = ((file_crc << 8) & 0xFF00) | ((file_crc >> 8) & 0x00FF)

        if (calculated_crc := int.from_bytes(calculate_crc16(file_data))) != swapped_crc:
            msg = (
                f"Computed CRC {calculated_crc:04x} for file {file_type} "
                f"does not match expected value {swapped_crc:04x}"
            )
            raise ReadException(msg)

        return file_data

    async def login(self, username: str, password: str) -> bool:
        """Login onto the inverter."""
        _LOGGER.debug("Logging in '%s'", username)
        # this circumvents the locking issue when using self.execute which locks on
        # _communication_lock in AsyncSmartTransport.send_and_receive
        assert isinstance(self.transport, AsyncSmartTransport)
        inverter_challenge = await self.transport.base_transport.send_and_receive(
            self.unit_id,
            LoginRequestChallengePDU(),
        )

        logged_in = await self.transport.base_transport.send_and_receive(
            self.unit_id,
            LoginPDU(username, password, inverter_challenge),
        )
        if logged_in:
            # Make sure we re-login after a reconnect
            assert isinstance(self.transport, AsyncSmartTransport)

            async def login_on_reconnect() -> None:
                """Login again after a reconnect."""
                _LOGGER.info("Reconnected to inverter, logging in again")
                logged_in_again = await self.login(username, password)
                if not logged_in_again:
                    _LOGGER.error("Failed to login after reconnect. Will not try again")
                    assert isinstance(self.transport, AsyncSmartTransport)
                    self.transport.on_reconnected = None
                else:
                    _LOGGER.info("Successfully logged in again after reconnect")

            async def login_on_reconnect_with_timeout() -> None:
                """Login again after a reconnect, with timeout."""
                return await asyncio.wait_for(login_on_reconnect(), timeout=WAIT_FOR_LOGIN_TIMEOUT)

            self.transport.on_reconnected = login_on_reconnect_with_timeout

        return logged_in

    async def heartbeat(self) -> bool:
        """Perform the heartbeat command. Only useful when maintaining a session."""
        if not self.connected:
            return False
        try:
            # 49999 is the magic register used to keep the connection alive
            await self.write_single_register(
                HEARTBEAT_REGISTER,
                0x1,
            )
        except ModbusResponseError as e:
            _LOGGER.warning("Received an error response when writing to the heartbeat register: %02x", e.error_code)
            return False
        except TModbusError:
            _LOGGER.exception("Exception during heartbeat")
            return False
        else:
            _LOGGER.debug("Heartbeat succeeded")
            return True


def create_client(
    transport: AsyncTcpTransport | AsyncRtuTransport,
    *,
    unit_id: int = DEFAULT_UNIT_ID,
    wait_after_connect: float = 1.0,
    wait_between_requests: float = DEFAULT_COOLDOWN_TIME,
) -> AsyncHuaweiSolarClient:
    """Create an AsyncHuaweiSolar instance."""
    # Wrap the transport in a smart transport to add auto-reconnect and cooldown between requests

    smart_transport = AsyncSmartTransport(
        transport,
        auto_reconnect=RECONNECT_RETRY_STRATEGY,
        wait_after_connect=wait_after_connect,
        wait_between_requests=wait_between_requests,
        response_retry_strategy=RESPONSE_RETRY_STRATEGY,
        retry_on_device_busy=True,
        retry_on_device_failure=True,
    )
    return AsyncHuaweiSolarClient(smart_transport, unit_id=unit_id)


def create_tcp_client(
    host: str,
    port: int = DEFAULT_TCP_PORT,
    *,
    unit_id: int = DEFAULT_UNIT_ID,
    timeout: int = DEFAULT_TIMEOUT,
    wait_after_connect: float = 1.0,
    wait_between_requests: float = DEFAULT_COOLDOWN_TIME,
) -> AsyncHuaweiSolarClient:
    """Create an AsyncHuaweiSolarClient connected via TCP."""
    transport = AsyncTcpTransport(host, port, timeout=timeout)
    return create_client(
        transport,
        unit_id=unit_id,
        wait_after_connect=wait_after_connect,
        wait_between_requests=wait_between_requests,
    )


def create_rtu_client(
    port: str,
    *,
    baudrate: int = DEFAULT_BAUDRATE,
    unit_id: int = DEFAULT_UNIT_ID,
    wait_after_connect: float = 1.0,
    wait_between_requests: float = DEFAULT_COOLDOWN_TIME,
) -> AsyncHuaweiSolarClient:
    """Create an AsyncHuaweiSolarClient connected via RTU."""
    transport = AsyncRtuTransport(port, baudrate=baudrate)
    return create_client(
        transport,
        unit_id=unit_id,
        wait_after_connect=wait_after_connect,
        wait_between_requests=wait_between_requests,
    )



"""Huawei SCharger device support."""

from huawei_solar import register_names as rn

from .base import HuaweiSolarDevice


class SChargerDevice(HuaweiSolarDevice):
    """An SCharger device."""

    software_version: str

    @classmethod
    def supports_device(cls, model_name: str) -> bool:
        """Check if this class support the given device."""
        return model_name.startswith("FusionCharge")

    async def _populate_additional_fields(self) -> None:
        (
            serial_number_result,
            software_version_result,
        ) = await self.client.get_multiple(
            [
                rn.CHARGER_ESN,
                rn.CHARGER_SOFTWARE_VERSION,
            ],
        )
        self.serial_number = serial_number_result.value
        self.software_version = software_version_result.value

        self.model_name = (await self.get(rn.CHARGER_MODEL)).value



"""Interact with Huawei inverters over Modbus."""

from .device import (
    EMMADevice,
    HuaweiSolarDevice,
    HuaweiSolarDeviceWithLogin,
    SChargerDevice,
    SDongleDevice,
    SmartLoggerDevice,
    SUN2000Device,
    create_device_instance,
    create_sub_device_instance,
)
from .device_discovery import DeviceIdentifier, DeviceInfo, get_device_identifiers, get_device_infos
from .exceptions import (
    ConnectionException,
    ConnectionInterruptedException,
    DecodeError,
    EncodeError,
    HuaweiSolarException,
    InvalidCredentials,
    PeakPeriodsValidationError,
    ReadException,
    TimeOfUsePeriodsException,
    WriteException,
)
from .files import (
    OptimizerHistoryRealTimeDataUnit,
    OptimizerOnlineStatus,
    OptimizerRealTimeData,
    OptimizerRunningStatus,
    OptimizerSystemInformation,
)
from .modbus_client import AsyncHuaweiSolarClient, create_rtu_client, create_tcp_client
from .register_definitions import Result
from .register_names import RegisterName

__all__ = [
    "AsyncHuaweiSolarClient",
    "ConnectionException",
    "ConnectionInterruptedException",
    "DecodeError",
    "DeviceIdentifier",
    "DeviceInfo",
    "EMMADevice",
    "EncodeError",
    "HuaweiSolarDevice",
    "HuaweiSolarDeviceWithLogin",
    "HuaweiSolarException",
    "InvalidCredentials",
    "OptimizerHistoryRealTimeDataUnit",
    "OptimizerOnlineStatus",
    "OptimizerRealTimeData",
    "OptimizerRunningStatus",
    "OptimizerSystemInformation",
    "PeakPeriodsValidationError",
    "ReadException",
    "RegisterName",
    "Result",
    "SChargerDevice",
    "SDongleDevice",
    "SUN2000Device",
    "SmartLoggerDevice",
    "TimeOfUsePeriodsException",
    "WriteException",
    "create_device_instance",
    "create_rtu_client",
    "create_sub_device_instance",
    "create_tcp_client",
    "get_device_identifiers",
    "get_device_infos",
]



"""Higher-level access to Huawei Solar inverters."""

import logging
from contextlib import suppress
from datetime import UTC, datetime, timedelta
from typing import Any

from huawei_solar import register_names as rn
from huawei_solar import register_values as rv
from huawei_solar.exceptions import (
    HuaweiSolarException,
    ReadException,
)
from huawei_solar.files import (
    OptimizerRealTimeData,
    OptimizerRealTimeDataFile,
    OptimizerSystemInformation,
    OptimizerSystemInformationDataFile,
)
from huawei_solar.register_definitions import Result, TimestampRegister
from huawei_solar.registers import METER_REGISTERS, REGISTERS

from .base import HuaweiSolarDeviceWithLogin
from .emma import EMMADevice
from .smartlogger import SmartLoggerDevice

_LOGGER = logging.getLogger(__name__)

MAX_NUMBER_OF_PV_STRINGS = 24


class SUN2000Device(HuaweiSolarDeviceWithLogin):
    """A Huawei SUN2000 device."""

    serial_number: str
    product_number: str
    firmware_version: str
    software_version: str

    pv_string_count: int = 0
    has_optimizers: bool = False

    battery_1_type: rv.StorageProductModel = rv.StorageProductModel.NONE
    battery_2_type: rv.StorageProductModel = rv.StorageProductModel.NONE
    supports_capacity_control = False
    power_meter_online = False
    power_meter_type: rv.MeterType | None = None

    _pv_registers: list[str]

    _time_zone: int | None = None
    _dst: bool | None = None

    _previous_device_status: str | None = None

    @classmethod
    def supports_device(cls, model_name: str) -> bool:
        """Check if this class support the given device."""
        return model_name.startswith(
            (
                "SUN",
                "EDF ESS",
                "Powershifter",
                "SWI300",
            ),
        )

    async def _populate_additional_fields(self) -> None:
        (
            serial_number_result,
            pn_result,
            firmware_version_result,
            software_version_result,
        ) = await self.client.get_multiple(
            [
                rn.SERIAL_NUMBER,
                rn.PN,
                rn.FIRMWARE_VERSION,
                rn.SOFTWARE_VERSION,
            ],
        )
        self.serial_number = serial_number_result.value
        self.product_number = pn_result.value
        self.firmware_version = firmware_version_result.value
        self.software_version = software_version_result.value

        self.pv_string_count = (await self.get(rn.NB_PV_STRINGS)).value
        self._pv_registers = _compute_pv_registers(self.pv_string_count)

        # some inverters throw an IllegalAddress exception when accessing this address
        with suppress(ReadException):
            self.has_optimizers = (await self.get(rn.NB_OPTIMIZERS)).value

        with suppress(ReadException):
            self.battery_1_type = (await self.get(rn.STORAGE_UNIT_1_PRODUCT_MODEL)).value

        with suppress(ReadException):
            self.battery_2_type = (await self.get(rn.STORAGE_UNIT_2_PRODUCT_MODEL)).value

        if (
            self.battery_1_type is not rv.StorageProductModel.NONE
            and self.battery_2_type is not rv.StorageProductModel.NONE
            and self.battery_1_type != self.battery_2_type
        ):
            _LOGGER.warning(
                "Detected two batteries of a different type. This can lead to unexpected behavior",
            )

        if self.battery_type != rv.StorageProductModel.NONE and (
            self.primary_device is None or not isinstance(self.primary_device, (EMMADevice, SmartLoggerDevice))
        ):
            try:
                await self.get(rn.STORAGE_CAPACITY_CONTROL_MODE)
                self.supports_capacity_control = True
            except ReadException:
                _LOGGER.debug("Storage capacity control as it is not supported by device %d", self.client.unit_id)
                self.supports_capacity_control = False

        with suppress(ReadException):
            self.power_meter_online = (await self.get(rn.METER_STATUS)).value == rv.MeterStatus.NORMAL

        # Caveat: if the inverter is in offline mode, and the power meter is thus offline,
        # we will incorrectly detect that no power meter is present.
        if self.power_meter_online:
            self.power_meter_type = (await self.get(rn.METER_TYPE)).value

        # reading these registers fails on some firmware versions (cfr. https://github.com/wlcrs/huawei_solar/issues/1264)
        with suppress(ReadException):
            self._dst = (await self.get(rn.DAYLIGHT_SAVING_TIME)).value
        with suppress(ReadException):
            self._time_zone = (await self.get(rn.TIME_ZONE)).value

    def _handle_batch_read_error(
        self,
        queried_register_names: list[rn.RegisterName],
        exc: HuaweiSolarException,
    ) -> None:
        """Handle read errors in batch_update."""
        if any(regname in METER_REGISTERS for regname in queried_register_names):
            _LOGGER.info(
                "Fetching power meter registers failed. "
                "We'll assume that this is due to the power meter going offline and the registers "
                "becoming invalid as a result",
                exc_info=exc,
            )
            self.power_meter_online = False

        raise exc

    def _detect_state_changes(self, new_values: dict[rn.RegisterName, Result[Any]]) -> None:
        """Update state based on result of batch_update query.

        Used by subclasses to detect important changes.
        """
        # When there is a power outage, but the installation stays online with a backup box installed,
        # then the power meter goes offline. If we still try to query it, the inverter will close the connection.
        # To prevent this, we always check if the power meter is still online when the device status changes.
        #
        # cfr. https://gitlab.com/Emilv2/huawei-solar/-/merge_requests/9#note_1281471842

        if rn.DEVICE_STATUS in new_values:
            new_device_status = new_values[rn.DEVICE_STATUS].value
            if self._previous_device_status != new_device_status:
                _LOGGER.debug(
                    "Detected a device state change from %s to %s : resetting power meter online status",
                    self._previous_device_status,
                    new_device_status,
                )
                self.power_meter_online = False

            self._previous_device_status = new_device_status

    async def _filter_registers(self, register_names: list[rn.RegisterName]) -> list[rn.RegisterName]:
        result = register_names

        # Filter out power meter registers if the power meter is offline
        power_meter_register_names = {rn for rn in register_names if rn in METER_REGISTERS}
        if power_meter_register_names:
            # Do a check of the METER_STATUS register only if the power meter is marked offline
            if not self.power_meter_online:
                power_meter_online_register = await self.get(rn.METER_STATUS)
                self.power_meter_online = power_meter_online_register.value

                _LOGGER.debug("Power meter online: %s", self.power_meter_online)

            # If it is still offline after the check then filter out all power meter registers
            if not self.power_meter_online:
                _LOGGER.debug(
                    "Removing power meter registers as the power meter is offline",
                )
                result = list(
                    filter(
                        lambda regname: regname == rn.METER_STATUS or rn not in power_meter_register_names,
                        register_names,
                    ),
                )

        return result

    def _transform_register_values(self, register_name: rn.RegisterName, result: Result[Any]) -> Result[Any]:
        if isinstance(REGISTERS[register_name], TimestampRegister) and result.value is not None:
            assert isinstance(result.value, datetime)
            value = result.value
            if self._time_zone:
                value -= timedelta(minutes=self._time_zone)
            # if DST is in effect, we need to shift another hour.
            if self._dst:
                value -= timedelta(hours=1)

            return Result(value.astimezone(tz=UTC), result.unit)

        return result

    async def _get_system_time(self) -> int | None:
        """Get the system time from the inverter."""
        if self.primary_device and isinstance(self.primary_device, EMMADevice):
            # Inverters don't return their own system time when connected via EMMA.
            # Instead, we need to read the system time from the EMMA device.

            return (await self.primary_device.get(rn.EMMA_SYSTEM_TIME)).value  # type: ignore[no-any-return]

        return (await self.get(rn.SYSTEM_TIME_RAW)).value  # type: ignore[no-any-return]

    async def get_latest_optimizer_history_data(
        self,
    ) -> dict[int, OptimizerRealTimeData]:
        """Read the latest Optimizer History Data File from the inverter."""
        # emulates behavior from FusionSolar app when current status of optimizers is queried
        end_time = await self._get_system_time()
        if end_time is None:
            msg = "Could not retrieve system time. Cannot proceed with reading optimizer data."
            raise ReadException(msg)
        start_time = end_time - 600

        file_data = await self.read_file(
            OptimizerRealTimeDataFile.FILE_TYPE,
            OptimizerRealTimeDataFile.query_within_timespan(start_time, end_time),
        )
        real_time_data = OptimizerRealTimeDataFile(file_data)

        if len(real_time_data.data_units) == 0:
            return {}

        # we only expect one element, but if more would be present,
        # then only the latest one is of interest (list is sorted time descending)
        latest_unit = real_time_data.data_units[0]

        return {opt.optimizer_address: opt for opt in latest_unit.optimizers}

    async def get_optimizer_system_information_data(
        self,
    ) -> dict[int, OptimizerSystemInformation]:
        """Read the Optimizer System Information Data File from the inverter."""
        file_data = await self.read_file(OptimizerSystemInformationDataFile.FILE_TYPE)
        system_information_data = OptimizerSystemInformationDataFile(file_data)

        return {opt.optimizer_address: opt for opt in system_information_data.optimizers}

    @property
    def battery_type(self) -> rv.StorageProductModel:
        """The battery type present on this inverter."""
        if self.battery_1_type != rv.StorageProductModel.NONE:
            return self.battery_1_type
        return self.battery_2_type


def _compute_pv_registers(pv_string_count: int) -> list[str]:
    """Get the registers for the PV strings which were detected from the inverter."""
    assert 1 <= pv_string_count <= MAX_NUMBER_OF_PV_STRINGS

    pv_registers = []
    for idx in range(1, pv_string_count + 1):
        pv_registers.extend(
            [
                getattr(rn, f"PV_{idx:02}_VOLTAGE"),
                getattr(rn, f"PV_{idx:02}_CURRENT"),
            ],
        )
    return pv_registers



"""String register definition."""

from typing import Any

from .base import RegisterDefinition, Result, TargetDevice


class StringRegister(RegisterDefinition[str]):
    """A string register."""

    def __init__(
        self,
        register: int,
        length: int,
        *,
        writeable: bool = False,
        readable: bool = True,
        target_device: TargetDevice = TargetDevice.SUN2000,
    ) -> None:
        """Create StringRegister."""
        super().__init__(
            register,
            writeable=writeable,
            readable=readable,
            target_device=target_device,
        )
        self.length = length
        self.format = f"{length * 2}s"  # every register has two bytes

    def encode(self, data: str) -> tuple[Any, ...]:
        """Encode string into registers."""
        return (data.encode("utf-8"),)

    def decode(self, values: tuple[Any, ...]) -> Result[str]:
        """Decode string."""
        return Result(bytes_to_string(values[0]), None)


def bytes_to_string(value: bytes) -> str:
    """Convert bytes to string, stripping any null bytes."""
    # Strip anything after the first null-byte, as that is typically garbage
    null_byte_index = value.find(b"\x00")
    if null_byte_index != -1:
        value = value[:null_byte_index]
    return value.decode("utf-8", errors="backslashreplace")



"""Custom PDU classes for tmodbus."""

import hmac
import logging
import secrets
import struct
from dataclasses import dataclass, field
from hashlib import sha256

from tmodbus.exceptions import ModbusResponseError, register_custom_exception
from tmodbus.pdu import BaseSubFunctionClientPDU, register_pdu_class

RECONNECT_DELAY = 1000  # in milliseconds
WAIT_ON_CONNECT = 1500  # in milliseconds

_LOGGER = logging.getLogger(__name__)


# Register custom Huawei Modbus PDUs
@dataclass(frozen=True)
class LoginRequestChallengePDU(BaseSubFunctionClientPDU[bytes]):
    """Modbus PDU to request a login challenge."""

    function_code = 0x41
    sub_function_code = 0x24
    rtu_byte_count_pos = 3

    def encode_request(self) -> bytes:
        """Encode LoginRequestChallengePDU."""
        data_length = 1
        value = 0
        return struct.pack(">BBBB", self.function_code, self.sub_function_code, data_length, value)

    def decode_response(self, response: bytes) -> bytes:
        """Decode LoginRequestChallengePDU response."""
        response_header_struct = struct.Struct(">BBB")
        (function_code, sub_function_code, response_content_length) = response_header_struct.unpack_from(response, 0)

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:02x}, received {function_code:02x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = (
                f"Unexpected sub function code: expected {self.sub_function_code:02x}, received {sub_function_code:02x}"
            )
            raise ValueError(msg)

        expected_response_content_length = 17
        if expected_response_content_length != response_content_length:
            msg = (
                f"Invalid response content length length: expected {expected_response_content_length}, "
                f"received {response_content_length}"
            )
            raise ValueError(msg)

        inverter_challenge_length = 16
        return response[response_header_struct.size : response_header_struct.size + inverter_challenge_length]


register_pdu_class(LoginRequestChallengePDU)


def _compute_digest(password: bytes, seed: bytes) -> bytes:
    hashed_password = sha256(password).digest()

    return hmac.digest(key=hashed_password, msg=seed, digest=sha256)


@dataclass(frozen=True)
class LoginPDU(BaseSubFunctionClientPDU[bool]):
    """Login PDU."""

    function_code = 0x41
    sub_function_code = 0x25
    rtu_byte_count_pos = 3

    username: str
    password: str
    inverter_challenge: bytes

    client_challenge: bytes = field(default_factory=lambda: secrets.token_bytes(16))

    def encode_request(self) -> bytes:
        """Encode the login request."""
        encoded_username = self.username.encode("utf-8")
        hashed_password = _compute_digest(
            self.password.encode("utf-8"),
            self.inverter_challenge,
        )

        total_length = len(self.client_challenge) + 1 + len(encoded_username) + 1 + len(hashed_password)

        return bytes(
            [
                self.function_code,
                self.sub_function_code,
                total_length,
                *self.client_challenge,
                len(encoded_username),
                *encoded_username,
                len(hashed_password),
                *hashed_password,
            ],
        )

    def decode_response(self, response: bytes) -> bool:
        """Decode LoginPDU response and check the returned MAC."""
        response_header_struct = struct.Struct(">BBB?B")
        (
            function_code,
            sub_function_code,
            _response_content_length,
            failure,
            inverter_mac_response_length,
        ) = response_header_struct.unpack_from(response, 0)

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:02x}, received {function_code:02x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = f"Invalid sub function code: expected {self.sub_function_code:02x}, received {sub_function_code:02x}"
            raise ValueError(msg)

        if failure:
            return False

        inverter_mac_response = response[
            response_header_struct.size : response_header_struct.size + inverter_mac_response_length
        ]

        if _compute_digest(self.password.encode("utf-8"), self.client_challenge) != inverter_mac_response:
            msg = "Inverter response contains an invalid challenge answer. This could indicate a MitM-attack!"
            raise ValueError(msg)

        return True


register_pdu_class(LoginPDU)


@dataclass(frozen=True, slots=True)
class StartFileUpload:
    """Contents of StartFileUpload response."""

    file_length: int
    data_frame_length: int
    customised_data: bytes


@dataclass(frozen=True)
class StartFileUploadPDU(BaseSubFunctionClientPDU[StartFileUpload]):
    """Modbus file upload request."""

    function_code = 0x41
    sub_function_code = 0x05

    file_type: int
    customised_data: bytes = field(default_factory=bytes)

    def encode_request(self) -> bytes:
        """Encode request."""
        data_length = 1 + len(self.customised_data)
        return (
            struct.pack(">BBBB", self.function_code, self.sub_function_code, data_length, self.file_type)
            + self.customised_data
        )

    def decode_response(self, response: bytes) -> StartFileUpload:
        """Decode response."""
        response_header_struct = struct.Struct(">BBBBLB")
        (
            function_code,
            sub_function_code,
            data_length,
            file_type,
            file_length,
            data_frame_length,
        ) = response_header_struct.unpack_from(response, 0)

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:02x}, received {function_code:02x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = f"Invalid sub function code: expected {self.sub_function_code:02x}, received {sub_function_code:02x}"
            raise ValueError(msg)

        customised_data = response[response_header_struct.size :]
        expected_data_length = struct.calcsize(">BLB") + len(customised_data)

        if data_length != expected_data_length:
            msg = f"Invalid data length: expected {expected_data_length}, received {data_length}"
            raise ValueError(msg)

        if file_type != self.file_type:
            msg = f"Invalid file type: expected {self.file_type:02x}, received {file_type:02x}"
            raise ValueError(msg)

        return StartFileUpload(
            file_length=file_length,
            data_frame_length=data_frame_length,
            customised_data=customised_data,
        )


register_pdu_class(StartFileUploadPDU)


@dataclass(frozen=True, slots=True)
class UploadFileFrame:
    """Represents a frame of file data."""

    frame_no: int
    frame_data: bytes


@dataclass(frozen=True)
class UploadFileFramePDU(BaseSubFunctionClientPDU[UploadFileFrame]):
    """Modbus Request for (a part of) a file."""

    function_code = 0x41
    rtu_byte_count_pos = 3

    sub_function_code = 0x06

    file_type: int
    frame_no: int

    def encode_request(self) -> bytes:
        """Encode UploadFileFramePDU."""
        data_length = 3
        return struct.pack(
            ">BBBBH",
            self.function_code,
            self.sub_function_code,
            data_length,
            self.file_type,
            self.frame_no,
        )

    def decode_response(self, response: bytes) -> UploadFileFrame:
        """Decode UploadPDU response."""
        response_header_struct = struct.Struct(">BBBBH")
        (
            function_code,
            sub_function_code,
            data_length,
            file_type,
            frame_no,
        ) = response_header_struct.unpack_from(response, 0)

        frame_data = response[response_header_struct.size :]

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:02x}, received {function_code:02x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = f"Invalid sub function code: expected {self.sub_function_code:02x}, received {sub_function_code:02x}"
            raise ValueError(msg)

        expected_data_length = struct.calcsize(">BH") + len(frame_data)
        if data_length != expected_data_length:
            msg = f"Invalid data length: expected {expected_data_length}, received {data_length}"
            raise ValueError(msg)

        if file_type != self.file_type:
            msg = f"Invalid file type: expected {self.file_type:02x}, received {file_type:02x}"
            raise ValueError(msg)

        return UploadFileFrame(frame_no=frame_no, frame_data=frame_data)


register_pdu_class(UploadFileFramePDU)


@dataclass(frozen=True)
class CompleteUploadPDU(BaseSubFunctionClientPDU[int]):
    """Modbus Request to complete a file upload.

    Returns the file CRC value.
    """

    function_code = 0x41
    rtu_byte_count_pos = 3

    sub_function_code = 0x0C

    file_type: int

    def encode_request(self) -> bytes:
        """Encode CompleteUploadModbusRequest."""
        data_length = 1
        return struct.pack(">BBBB", self.function_code, self.sub_function_code, data_length, self.file_type)

    def decode_response(self, response: bytes) -> int:
        """Decode CompleteUploadModbusResponse."""
        file_crc: int
        function_code, sub_function_code, data_length, file_type, file_crc = struct.unpack(">BBBBH", response)

        if function_code != self.function_code:
            msg = f"Invalid function code: expected {self.function_code:02x}, received {function_code:02x}"
            raise ValueError(msg)

        if sub_function_code != self.sub_function_code:
            msg = f"Invalid sub function code: expected {self.sub_function_code:02x}, received {sub_function_code:02x}"
            raise ValueError(msg)

        expected_data_length = 3
        if data_length != expected_data_length:
            msg = f"Invalid data length: expected {expected_data_length}, received {data_length}"
            raise ValueError(msg)

        if file_type != self.file_type:
            msg = f"Invalid file type: expected {self.file_type:02x}, received {file_type:02x}"
            raise ValueError(msg)

        return file_crc


register_pdu_class(CompleteUploadPDU)


class PermissionDeniedError(ModbusResponseError):
    """Permission Denied exception.

    Raised when the device returns a permission denied error.
    """

    error_code = 0x80


register_custom_exception(PermissionDeniedError)



"""Register definitions of the complex registers containing time-periods."""

from dataclasses import dataclass
from enum import IntEnum
from typing import Any

from huawei_solar.exceptions import PeakPeriodsValidationError, TimeOfUsePeriodsException

from .base import RegisterDefinition, Result


@dataclass(frozen=True, slots=True)
class LG_RESU_TimeOfUsePeriod:
    """Time of use period of LG RESU."""

    start_time: int  # minutes since midnight
    end_time: int  # minutes since midnight
    electricity_price: float


class ChargeFlag(IntEnum):
    """Charge Flag."""

    CHARGE = 0
    DISCHARGE = 1


@dataclass(frozen=True, slots=True)
class HUAWEI_LUNA2000_TimeOfUsePeriod:
    """Time of use period of Huawei LUNA2000."""

    start_time: int  # minutes since midnight
    end_time: int  # minutes since midnight
    charge_flag: ChargeFlag
    days_effective: tuple[
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
    ]  # Valid on days Sunday to Saturday


LG_RESU_TOU_PERIODS = 10


class LG_RESU_TimeOfUseRegisters(RegisterDefinition[list[LG_RESU_TimeOfUsePeriod]]):
    """Time of use register."""

    format = f"H{'HHI' * LG_RESU_TOU_PERIODS}"
    format_size = 1 + 3 * LG_RESU_TOU_PERIODS
    length = 41

    def decode(self, values: tuple[Any, ...]) -> Result[list[LG_RESU_TimeOfUsePeriod]]:
        """Decode time of use register."""
        number_of_periods = values[0]
        assert number_of_periods <= LG_RESU_TOU_PERIODS

        def _decode_lg_resu_tou_period(
            start_time: int,
            end_time: int,
            electricity_price: int,
        ) -> LG_RESU_TimeOfUsePeriod:
            return LG_RESU_TimeOfUsePeriod(
                start_time,
                end_time,
                electricity_price / 1000,
            )

        return Result(
            value=[
                _decode_lg_resu_tou_period(*values[1 + idx * 3 : 1 + ((idx + 1) * 3)])
                for idx in range(number_of_periods)
            ],
            unit=None,
        )

    def _validate(
        self,
        data: list[LG_RESU_TimeOfUsePeriod],
    ) -> None:
        """Validate data type."""
        if len(data) == 0:
            return  # nothing to check

        # Sanity check of each period individually
        for tou_period in data:
            if tou_period.start_time < 0 or tou_period.end_time < 0:
                msg = "TOU period is invalid (Below zero)"
                raise TimeOfUsePeriodsException(msg)
            if tou_period.start_time > 24 * 60 or tou_period.end_time > 24 * 60:
                msg = "TOU period is invalid (Spans over more than one day)"
                raise TimeOfUsePeriodsException(
                    msg,
                )
            if tou_period.start_time >= tou_period.end_time:
                msg = "TOU period is invalid (start-time is greater than end-time)"
                raise TimeOfUsePeriodsException(
                    msg,
                )

        # make a copy of the data to sort
        sorted_periods: list[LG_RESU_TimeOfUsePeriod] = data.copy()

        sorted_periods.sort(key=lambda a: a.start_time)

        for period_idx in range(1, len(sorted_periods)):
            current_period = sorted_periods[period_idx]
            prev_period = sorted_periods[period_idx - 1]
            if (
                prev_period.start_time <= current_period.start_time < prev_period.end_time
                or prev_period.start_time < current_period.end_time <= prev_period.end_time
            ):
                msg = "TOU periods are overlapping"
                raise TimeOfUsePeriodsException(msg)

    def encode(self, data: list[LG_RESU_TimeOfUsePeriod]) -> tuple[Any, ...]:
        """Encode Time Of Use Period registers."""
        self._validate(data)

        values = [len(data)]

        for period in data:
            values.extend(
                [
                    period.start_time,
                    period.end_time,
                    int(period.electricity_price * 1000),
                ],
            )

        values.extend([0, 0, 0] * (LG_RESU_TOU_PERIODS - len(data)))

        return tuple(values)


HUAWEI_LUNA2000_TOU_PERIODS = 14


class HUAWEI_LUNA2000_TimeOfUseRegisters(RegisterDefinition[list[HUAWEI_LUNA2000_TimeOfUsePeriod]]):
    """Time of use register."""

    format = f"H{'HHBB' * HUAWEI_LUNA2000_TOU_PERIODS}"
    format_size = 1 + 4 * HUAWEI_LUNA2000_TOU_PERIODS
    length = 43

    def decode(self, values: tuple[Any, ...]) -> Result[list[HUAWEI_LUNA2000_TimeOfUsePeriod]]:
        """Decode time of use register."""
        number_of_periods = values[0]
        assert number_of_periods <= HUAWEI_LUNA2000_TOU_PERIODS

        def _decode_huawei_luna2000_tou_period(
            start_time: int,
            end_time: int,
            charge: int,
            days_effective: int,
        ) -> HUAWEI_LUNA2000_TimeOfUsePeriod:
            return HUAWEI_LUNA2000_TimeOfUsePeriod(
                start_time,
                end_time,
                ChargeFlag(charge),
                _days_effective_parser(days_effective),
            )

        return Result(
            [
                _decode_huawei_luna2000_tou_period(*values[1 + idx * 4 : 1 + ((idx + 1) * 4)])
                for idx in range(number_of_periods)
            ],
            unit=None,
        )

    def _validate(
        self,
        data: list[HUAWEI_LUNA2000_TimeOfUsePeriod],
    ) -> None:
        """Validate data type."""
        if len(data) == 0:
            return  # nothing to check

        # Sanity check of each period individually
        for tou_period in data:
            if not isinstance(tou_period, HUAWEI_LUNA2000_TimeOfUsePeriod):
                msg = "TOU period is of an unexpected type"
                raise TimeOfUsePeriodsException(msg)
            if tou_period.start_time < 0 or tou_period.end_time < 0:
                msg = "TOU period is invalid (Below zero)"
                raise TimeOfUsePeriodsException(msg)
            if tou_period.start_time > 24 * 60 or tou_period.end_time > 24 * 60:
                msg = "TOU period is invalid (Spans over more than one day)"
                raise TimeOfUsePeriodsException(
                    msg,
                )
            if tou_period.start_time >= tou_period.end_time:
                msg = "TOU period is invalid (start-time is greater than end-time)"
                raise TimeOfUsePeriodsException(
                    msg,
                )

        for day_idx in range(7):
            # find all ranges that are valid for the given day
            active_periods: list[HUAWEI_LUNA2000_TimeOfUsePeriod] = list(
                filter(lambda period: period.days_effective[day_idx], data),
            )

            active_periods.sort(key=lambda a: a.start_time)

            for period_idx in range(1, len(active_periods)):
                current_period = active_periods[period_idx]
                prev_period = active_periods[period_idx - 1]
                if (
                    prev_period.start_time <= current_period.start_time < prev_period.end_time
                    or prev_period.start_time < current_period.end_time <= prev_period.end_time
                ):
                    msg = "TOU periods are overlapping"
                    raise TimeOfUsePeriodsException(msg)

    def encode(
        self,
        data: list[HUAWEI_LUNA2000_TimeOfUsePeriod],
    ) -> tuple[Any, ...]:
        """Encode Time Of Use Period registers."""
        self._validate(data)

        assert len(data) <= HUAWEI_LUNA2000_TOU_PERIODS

        values = [len(data)]

        def _days_effective_builder(days_tuple: tuple[bool, bool, bool, bool, bool, bool, bool]) -> int:
            result = 0
            mask = 0x1
            for i in range(7):
                if days_tuple[i]:
                    result += mask
                mask = mask << 1

            return result

        for period in data:
            values.extend(
                [
                    period.start_time,
                    period.end_time,
                    int(period.charge_flag),
                    _days_effective_builder(period.days_effective),
                ],
            )

        values.extend([0, 0, 0, 0] * (HUAWEI_LUNA2000_TOU_PERIODS - len(data)))

        return tuple(values)


@dataclass(frozen=True, slots=True)
class ChargeDischargePeriod:
    """Charge or Discharge Period."""

    start_time: int  # minutes since midnight
    end_time: int  # minutes since midnight
    power: int  # power in watts


CHARGE_DISCHARGE_PERIODS = 10


class ChargeDischargePeriodRegisters(RegisterDefinition[list[ChargeDischargePeriod]]):
    """Charge or discharge period registers."""

    format = f"H{'HHI' * CHARGE_DISCHARGE_PERIODS}"
    format_size = 1 + 3 * CHARGE_DISCHARGE_PERIODS
    length = 41

    def decode(self, values: tuple[Any, ...]) -> Result[list[ChargeDischargePeriod]]:
        """Decode ChargeDischargePeriodRegisters."""
        number_of_periods = values[0]
        assert number_of_periods <= CHARGE_DISCHARGE_PERIODS

        def _decode_charge_discharge_period(start_time: int, end_time: int, power: int) -> ChargeDischargePeriod:
            return ChargeDischargePeriod(start_time, end_time, power)

        periods = [
            _decode_charge_discharge_period(*values[1 + idx * 3 : 1 + ((idx + 1) * 3)])
            for idx in range(number_of_periods)
        ]

        return Result(periods[:number_of_periods], unit=None)

    def encode(self, data: list[ChargeDischargePeriod]) -> tuple[Any, ...]:
        """Encode ChargeDischargePeriodRegisters."""
        assert len(data) <= CHARGE_DISCHARGE_PERIODS

        values = [len(data)]

        for period in data:
            values.extend(
                [
                    period.start_time,
                    period.end_time,
                    period.power,
                ],
            )

        # pad with empty periods
        values.extend([0, 0, 0] * (CHARGE_DISCHARGE_PERIODS - len(data)))

        return tuple(values)


@dataclass(frozen=True, slots=True)
class PeakSettingPeriod:
    """Peak Setting Period."""

    start_time: int  # minutes since midnight
    end_time: int  # minutes since midnight
    power: int  # power in watts
    days_effective: tuple[
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
        bool,
    ]  # Valid on days Sunday to


PEAK_SETTING_PERIODS = 14


def _days_effective_builder(days_tuple: tuple[bool, bool, bool, bool, bool, bool, bool]) -> int:
    result = 0
    mask = 0x1
    for i in range(7):
        if days_tuple[i]:
            result += mask
        mask = mask << 1

    return result


def _days_effective_parser(value: int) -> tuple[bool, bool, bool, bool, bool, bool, bool]:
    result = []
    mask = 0x1
    for _ in range(7):
        result.append((value & mask) != 0)
        mask = mask << 1

    return tuple(result)  # type: ignore[return-value]


class PeakSettingPeriodRegisters(RegisterDefinition[list[PeakSettingPeriod]]):
    """Peak Setting Period registers."""

    format = f"H{'HHIB' * PEAK_SETTING_PERIODS}"
    format_size = 1 + 4 * PEAK_SETTING_PERIODS
    length = 64

    def decode(self, values: tuple[Any, ...]) -> Result[list[PeakSettingPeriod]]:
        """Decode PeakSettingPeriodRegisters."""
        number_of_periods = values[0]

        # Safety check
        number_of_periods = min(number_of_periods, PEAK_SETTING_PERIODS)

        periods = []
        for idx in range(number_of_periods):
            start_time, end_time, peak_value, week_value = values[1 + (idx * 4) : 1 + ((idx + 1) * 4)]

            if start_time != end_time and week_value != 0:
                periods.append(
                    PeakSettingPeriod(
                        start_time,
                        end_time,
                        peak_value,
                        _days_effective_parser(week_value),
                    ),
                )

        return Result(periods[:number_of_periods], unit=None)

    def _validate(self, data: list[PeakSettingPeriod]) -> None:
        for day_idx in range(7):
            # find all ranges that are valid for the given day
            active_periods: list[PeakSettingPeriod] = list(
                filter(lambda period: period.days_effective[day_idx], data),
            )

            if not active_periods:
                msg = "All days of the week need to be covered"
                raise PeakPeriodsValidationError(
                    msg,
                )

            # require full day to be covered
            active_periods.sort(key=lambda a: a.start_time)

            if active_periods[0].start_time != 0:
                msg = "Every day must be covered from 00:00"
                raise PeakPeriodsValidationError(msg)

            for period_idx in range(1, len(active_periods)):
                current_period = active_periods[period_idx]
                prev_period = active_periods[period_idx - 1]
                if current_period.start_time not in (
                    prev_period.end_time,
                    prev_period.end_time + 1,
                ):
                    msg = "All moments of each day need to be covered"
                    raise PeakPeriodsValidationError(
                        msg,
                    )

            if active_periods[-1].end_time not in ((24 * 60) - 1, 24 * 60):
                msg = "Every day must be covered until 23:59"
                raise PeakPeriodsValidationError(
                    msg,
                )

    def encode(self, data: list[PeakSettingPeriod]) -> tuple[Any, ...]:
        """Encode PeakSettingPeriodRegisters."""
        if len(data) > PEAK_SETTING_PERIODS:
            data = data[:PEAK_SETTING_PERIODS]

        values = [len(data)]

        for period in data:
            values.extend(
                [
                    period.start_time,
                    period.end_time,
                    period.power,
                    _days_effective_builder(period.days_effective),
                ],
            )

        # pad with empty periods
        values.extend([0, 0, 0, 0] * (PEAK_SETTING_PERIODS - len(data)))

        return tuple(values)



"""Definitions of the devices supported by this library."""

from logging import getLogger

from tmodbus.exceptions import IllegalDataAddressError

from huawei_solar import register_names as rn
from huawei_solar.exceptions import DeviceDetectionError
from huawei_solar.modbus_client import AsyncHuaweiSolarClient

from .base import HuaweiSolarDevice, HuaweiSolarDeviceWithLogin
from .emma import EMMADevice
from .scharger import SChargerDevice
from .sdongle import SDongleDevice
from .smartlogger import SmartLoggerDevice
from .sun2000 import SUN2000Device

_LOGGER = getLogger(__name__)

DEFAULT_SDONGLE_UNIT_ID = 100


def get_device_class_for_model(model_name: str) -> type[HuaweiSolarDevice]:
    """Get the device class for the given model name."""
    for candidate_bridge_class in [SUN2000Device, EMMADevice, SChargerDevice, SDongleDevice, SmartLoggerDevice]:
        if candidate_bridge_class.supports_device(model_name):
            return candidate_bridge_class

    _LOGGER.warning("Unknown product model '%s'. Defaulting to a SUN2000 device.", model_name)

    # Default to SUN2000Bridge if no specific match is found
    return SUN2000Device


async def detect_device_type(client: AsyncHuaweiSolarClient) -> tuple[type[HuaweiSolarDevice], str]:
    """Detect the type of the connected device."""

    async def _detect_sdongle() -> bool:
        try:
            device_search_status = (await client.get(rn.SDONGLE_DEVICE_SEARCH_STATUS)).value
        except IllegalDataAddressError:
            _LOGGER.warning("Failed to detect device type for unit ID %d.", DEFAULT_SDONGLE_UNIT_ID)
            return False
        else:
            _LOGGER.debug(
                "Successfully retrieved SDongle 'device search status' register for unit ID %d: %s",
                DEFAULT_SDONGLE_UNIT_ID,
                device_search_status,
            )
            return True

    # Unit ID 100 is typically used by an SDongle. Fast track checking for that.
    if client.unit_id == DEFAULT_SDONGLE_UNIT_ID and await _detect_sdongle():
        return SDongleDevice, "SDongle"

    try:
        model_name: str = (await client.get(rn.MODEL_NAME)).value
    except IllegalDataAddressError:
        _LOGGER.info("MODEL_NAME is an illegal data address for unit ID %d.", client.unit_id)
    else:
        return get_device_class_for_model(model_name), model_name

    try:
        # The SmartLogger does not have a MODEL_NAME register, so we need to detect it differently
        smartlogger_device_name: str = (await client.get(rn.SMARTLOGGER_DEVICE_NAME)).value
    except IllegalDataAddressError:
        _LOGGER.info("SMARTLOGGER_DEVICE_NAME is an illegal data address for unit ID %d.", client.unit_id)
    else:
        return get_device_class_for_model(smartlogger_device_name), smartlogger_device_name

    if await _detect_sdongle():
        return SDongleDevice, "SDongle"

    # If we reach here, we couldn't detect the device type
    _LOGGER.warning("Failed to detect device type.")
    msg = "Unable to detect the device type. The device may not be supported or may not be responding correctly."
    raise DeviceDetectionError(msg)


async def create_device_instance(client: AsyncHuaweiSolarClient) -> HuaweiSolarDevice:
    """Detect the connected device and create the appropriate instance."""
    device_type, model_name = await detect_device_type(client)
    return await device_type.create(
        client,
        model_name=model_name,
        primary_device=None,  # we are creating the primary device!
    )


async def create_sub_device_instance(
    primary_device: HuaweiSolarDevice,
    unit_id: int,
) -> HuaweiSolarDevice:
    """Create a HuaweiSolarDevice instance for extra servers accessible as subdevices via an existing device."""
    if primary_device.client.unit_id == unit_id:
        msg = "The unit_id for the sub-device must be different from the primary device's unit_id."
        raise ValueError(msg)

    sub_client = primary_device.client.for_unit_id(unit_id)
    device_type, model_name = await detect_device_type(sub_client)
    return await device_type.create(
        sub_client,
        model_name=model_name,
        primary_device=primary_device,
    )


__all__ = [
    "EMMADevice",
    "HuaweiSolarDevice",
    "HuaweiSolarDeviceWithLogin",
    "SChargerDevice",
    "SDongleDevice",
    "SUN2000Device",
    "SmartLoggerDevice",
    "create_device_instance",
    "create_sub_device_instance",
]



"""Tests for peak period register validation and encoding/decoding."""

import huawei_solar.register_names as rn
import pytest
from huawei_solar.exceptions import PeakPeriodsValidationError
from huawei_solar.register_definitions.periods import PeakSettingPeriod
from huawei_solar.registers import REGISTERS

ppr = REGISTERS[rn.STORAGE_CAPACITY_CONTROL_PERIODS]


def test_simple() -> None:
    pp_valid = [
        PeakSettingPeriod(
            start_time=0,
            end_time=1440,
            power=2500,
            days_effective=(True, True, True, True, True, True, True),
        ),
    ]

    ppr._validate(pp_valid)


def test_invalid_start_time() -> None:
    pp = [
        PeakSettingPeriod(
            start_time=60 * 24 + 1,
            end_time=15,
            power=2500,
            days_effective=(True, True, True, True, True, True, True),
        ),
    ]

    with pytest.raises(
        expected_exception=PeakPeriodsValidationError,
        match="Every day must be covered from 00:00",
    ):
        ppr._validate(pp)


def test_invalid_end_time() -> None:
    pp = [
        PeakSettingPeriod(
            start_time=0,
            end_time=15,
            power=2500,
            days_effective=(True, True, True, True, True, True, True),
        ),
    ]

    with pytest.raises(
        expected_exception=PeakPeriodsValidationError,
        match="Every day must be covered until 23:59",
    ):
        ppr._validate(pp)

    pp2 = [
        PeakSettingPeriod(
            start_time=0,
            end_time=1441,
            power=2500,
            days_effective=(True, True, True, True, True, True, True),
        ),
    ]

    with pytest.raises(
        expected_exception=PeakPeriodsValidationError,
        match="Every day must be covered until 23:59",
    ):
        ppr._validate(pp2)


def test_all_days_of_week_covered() -> None:
    pp = [
        PeakSettingPeriod(
            start_time=0,
            end_time=15,
            power=2500,
            days_effective=(False, True, True, True, True, True, True),
        ),
    ]

    with pytest.raises(
        expected_exception=PeakPeriodsValidationError,
        match="All days of the week need to be covered",
    ):
        ppr._validate(pp)


def test_multiple_periods_on_a_day() -> None:
    pp = [
        PeakSettingPeriod(
            start_time=0,
            end_time=1439,
            power=2500,
            days_effective=(False, True, True, True, True, True, True),
        ),
        PeakSettingPeriod(
            start_time=0,
            end_time=600,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
        PeakSettingPeriod(
            start_time=600,
            end_time=1439,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
    ]

    ppr._validate(pp)

    encoded = ppr.encode(pp)
    assert ppr.decode(encoded).value == pp

    pp2 = [
        PeakSettingPeriod(
            start_time=0,
            end_time=1439,
            power=2500,
            days_effective=(False, True, True, True, True, True, True),
        ),
        PeakSettingPeriod(
            start_time=0,
            end_time=600,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
        PeakSettingPeriod(
            start_time=601,
            end_time=1439,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
    ]

    ppr._validate(pp2)

    pp3 = [
        PeakSettingPeriod(
            start_time=0,
            end_time=1439,
            power=2500,
            days_effective=(False, True, True, True, True, True, True),
        ),
        PeakSettingPeriod(
            start_time=0,
            end_time=600,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
        PeakSettingPeriod(
            start_time=602,
            end_time=1439,
            power=2500,
            days_effective=(True, False, False, False, False, False, False),
        ),
    ]

    with pytest.raises(
        expected_exception=PeakPeriodsValidationError,
        match="All moments of each day need to be covered",
    ):
        ppr._validate(pp3)


def test_capacity_control_register() -> None:
    value = [
        PeakSettingPeriod(0, 1439, 2551, (True, True, True, True, True, True, False)),
        PeakSettingPeriod(
            0,
            200,
            2550,
            (False, False, False, False, False, False, True),
        ),
        PeakSettingPeriod(
            200,
            1439,
            2449,
            (False, False, False, False, False, False, True),
        ),
    ]

    pspr = REGISTERS[rn.STORAGE_CAPACITY_CONTROL_PERIODS]

    payload = pspr.encode(value)

    decoded_result = pspr.decode(payload).value

    assert decoded_result == value



"""Tests for device detection in huawei_solar.device.__init__."""

from types import SimpleNamespace
from typing import Any
from unittest.mock import AsyncMock, Mock

import pytest
from huawei_solar.device.emma import EMMADevice
from huawei_solar.device.scharger import SChargerDevice
from huawei_solar.device.sdongle import SDongleDevice
from huawei_solar.device.smartlogger import SmartLoggerDevice
from huawei_solar.device.sun2000 import SUN2000Device
from huawei_solar.exceptions import DeviceDetectionError
from tmodbus.const import FunctionCode
from tmodbus.exceptions import IllegalDataAddressError

from huawei_solar import register_names as rn
from huawei_solar.device import DEFAULT_SDONGLE_UNIT_ID, detect_device_type, get_device_class_for_model


def _value_result(value: str) -> SimpleNamespace:
    return SimpleNamespace(value=value)


def _client_with_get(unit_id: int, side_effect: Any) -> Mock:  # noqa: ANN401
    client = Mock()
    client.unit_id = unit_id
    client.get = AsyncMock(side_effect=side_effect)
    return client


@pytest.fixture
def patched_supports_device(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(SUN2000Device, "supports_device", staticmethod(lambda model: model == "sun2000-model"))
    monkeypatch.setattr(EMMADevice, "supports_device", staticmethod(lambda model: model == "emma-model"))
    monkeypatch.setattr(SChargerDevice, "supports_device", staticmethod(lambda model: model == "scharger-model"))
    monkeypatch.setattr(SDongleDevice, "supports_device", staticmethod(lambda model: model == "sdongle-model"))
    monkeypatch.setattr(
        SmartLoggerDevice,
        "supports_device",
        staticmethod(lambda model: model == "smartlogger-model"),
    )


@pytest.mark.parametrize(
    ("model_name", "expected_class"),
    [
        ("sun2000-model", SUN2000Device),
        ("emma-model", EMMADevice),
        ("scharger-model", SChargerDevice),
        ("sdongle-model", SDongleDevice),
        ("smartlogger-model", SmartLoggerDevice),
    ],
)
def test_get_device_class_for_model_all_supported_types(
    patched_supports_device: None,
    model_name: str,
    expected_class: type,
) -> None:
    assert get_device_class_for_model(model_name) is expected_class


def test_get_device_class_for_model_unknown_defaults_to_sun2000(patched_supports_device: None) -> None:
    assert get_device_class_for_model("unknown-model") is SUN2000Device


@pytest.mark.parametrize(
    ("model_name", "expected_class"),
    [
        ("sun2000-model", SUN2000Device),
        ("emma-model", EMMADevice),
        ("scharger-model", SChargerDevice),
    ],
)
async def test_detect_device_type_from_model_name(
    patched_supports_device: None,
    model_name: str,
    expected_class: type,
) -> None:
    def side_effect(register: str) -> Any:  # noqa: ANN401
        if register == rn.MODEL_NAME:
            return _value_result(model_name)
        msg = f"Unexpected register read: {register!r}"
        raise AssertionError(msg)

    client = _client_with_get(unit_id=1, side_effect=side_effect)

    detected_class, detected_name = await detect_device_type(client)

    assert detected_class is expected_class
    assert detected_name == model_name


async def test_detect_device_type_smartlogger_when_model_name_illegal(
    patched_supports_device: None,
) -> None:
    def side_effect(register: str) -> Any:  # noqa: ANN401
        if register == rn.MODEL_NAME:
            raise IllegalDataAddressError(
                error_code=IllegalDataAddressError.error_code,
                function_code=FunctionCode.READ_HOLDING_REGISTERS,
            )
        if register == rn.SMARTLOGGER_DEVICE_NAME:
            return _value_result("smartlogger-model")
        msg = f"Unexpected register read: {register!r}"
        raise AssertionError(msg)

    client = _client_with_get(unit_id=1, side_effect=side_effect)

    detected_class, detected_name = await detect_device_type(client)

    assert detected_class is SmartLoggerDevice
    assert detected_name == "smartlogger-model"


async def test_detect_device_type_sdongle_fast_track_on_unit_100() -> None:
    def side_effect(register: str) -> Any:  # noqa: ANN401
        if register == rn.SDONGLE_DEVICE_SEARCH_STATUS:
            return _value_result("done")
        msg = f"Unexpected register read: {register!r}"
        raise AssertionError(msg)

    client = _client_with_get(unit_id=DEFAULT_SDONGLE_UNIT_ID, side_effect=side_effect)

    detected_class, detected_name = await detect_device_type(client)

    assert detected_class is SDongleDevice
    assert detected_name == "SDongle"


async def test_detect_device_type_sdongle_fallback_when_other_registers_illegal() -> None:
    def side_effect(register: str) -> Any:  # noqa: ANN401
        if register in (rn.MODEL_NAME, rn.SMARTLOGGER_DEVICE_NAME):
            raise IllegalDataAddressError(
                error_code=IllegalDataAddressError.error_code,
                function_code=FunctionCode.READ_HOLDING_REGISTERS,
            )
        if register == rn.SDONGLE_DEVICE_SEARCH_STATUS:
            return _value_result("done")
        msg = f"Unexpected register read: {register!r}"
        raise AssertionError(msg)

    client = _client_with_get(unit_id=1, side_effect=side_effect)

    detected_class, detected_name = await detect_device_type(client)

    assert detected_class is SDongleDevice
    assert detected_name == "SDongle"


async def test_detect_device_type_raises_when_no_detection_path_matches() -> None:
    def side_effect(register: str) -> Any:  # noqa: ANN401
        if register in (rn.MODEL_NAME, rn.SMARTLOGGER_DEVICE_NAME, rn.SDONGLE_DEVICE_SEARCH_STATUS):
            raise IllegalDataAddressError(
                error_code=IllegalDataAddressError.error_code,
                function_code=FunctionCode.READ_HOLDING_REGISTERS,
            )
        msg = f"Unexpected register read: {register!r}"
        raise AssertionError(msg)

    client = _client_with_get(unit_id=1, side_effect=side_effect)

    with pytest.raises(DeviceDetectionError, match="Unable to detect the device type"):
        await detect_device_type(client)



"""A Modbus client that knows about registers as defined in this library."""

import logging
import struct
from typing import TYPE_CHECKING, Any

from tmodbus.client import AsyncModbusClient
from tmodbus.exceptions import IllegalDataAddressError, ModbusConnectionError, ModbusResponseError, TModbusError

from huawei_solar import register_names as rn
from huawei_solar.const import MAX_BATCHED_REGISTERS_COUNT
from huawei_solar.exceptions import (
    ConnectionInterruptedException,
    ReadException,
    WriteException,
)
from huawei_solar.modbus_pdu import PermissionDeniedError
from huawei_solar.registers import REGISTERS

if TYPE_CHECKING:
    from huawei_solar.register_definitions import RegisterDefinition, Result


_LOGGER = logging.getLogger(__name__)


class RegisterAwareModbusClient(AsyncModbusClient):
    """A Modbus client that knows about registers."""

    def _get_register_definitions(self, names: list[rn.RegisterName]) -> "list[RegisterDefinition[Any]]":
        """Get register definitions by name."""
        unknown_register_names = set(names) - REGISTERS.keys()
        if unknown_register_names:
            msg = f"Did not recognize register names: {', '.join(unknown_register_names)}"
            raise ValueError(msg)

        return [REGISTERS[name] for name in names]

    def _validate_registers_readable(
        self,
        names: list[rn.RegisterName],
        registers: "list[RegisterDefinition[Any]]",
    ) -> None:
        """Validate whether the requested registers are readable."""
        unreadable_register_names = [
            register_name for register, register_name in zip(registers, names, strict=False) if not register.readable
        ]
        if unreadable_register_names:
            msg = f"Trying to read unreadable registers: {', '.join(unreadable_register_names)}"
            raise ValueError(msg)

    def _construct_struct_format(self, registers: "list[RegisterDefinition[Any]]") -> str:
        """Construct a struct format to interpret the registers content with."""
        struct_format = f">{registers[0].format}"
        for idx in range(1, len(registers)):
            register_distance = registers[idx].register - (registers[idx - 1].register + registers[idx - 1].length)
            if register_distance < 0:
                msg = (
                    f"Requested registers must be in monotonically increasing order, "
                    f"but {registers[idx - 1].register} + {registers[idx - 1].length} > {registers[idx].register}!"
                )
                raise ValueError(msg)

            if register_distance > MAX_BATCHED_REGISTERS_COUNT:
                msg = "Gap between requested registers is too large. Split it in two requests"
                raise ValueError(msg)

            struct_format += f"{'x' * 2 * register_distance}{registers[idx].format}"

        return struct_format

    def _decode_response_tuple(
        self,
        registers: "list[RegisterDefinition[Any]]",
        response: tuple[Any, ...],
    ) -> "list[Result[Any]]":
        """Decode response tuple."""
        result = []
        tuple_idx = 0
        for register in registers:
            register_values = register.decode(response[tuple_idx : tuple_idx + register.format_size])
            result.append(register_values)
            tuple_idx += register.format_size

        return result

    async def get_multiple(self, names: list[rn.RegisterName]) -> "list[Result[Any]]":
        """Read multiple registers at the same time.

        This is only possible if the registers are consecutively available in the
        inverters' memory.
        """
        if len(names) == 0:
            msg = "Expected at least one register name"
            raise ValueError(msg)

        registers = self._get_register_definitions(names)
        self._validate_registers_readable(names, registers)

        start_address = registers[0].register
        struct_format = self._construct_struct_format(registers)
        try:
            response_tuple = await self.read_struct_format(start_address, format_struct=struct_format)
        except ModbusResponseError as err:
            msg = f"Failed to read registers {', '.join(names)}: received {type(err).__name__}"
            raise ReadException(msg, modbus_exception_code=err.error_code) from err
        except ModbusConnectionError as err:
            _LOGGER.exception("Connection error while reading registers %s", names)
            msg = f"Connection failed when trying to read registers {', '.join(names)}"
            raise ConnectionInterruptedException(msg) from err
        except TModbusError as err:
            msg = f"Failed to read registers {', '.join(names)}: {err}"
            raise ReadException(msg) from err
        else:
            return self._decode_response_tuple(registers, response_tuple)

    async def get_multiple_as_dict(self, names: list[rn.RegisterName]) -> "dict[rn.RegisterName, Result[Any]]":
        """Read multiple registers and return them as a dictionary.

        This is only possible if the registers are consecutively available in the
        inverters' memory.
        """
        return dict(
            zip(
                names,
                await self.get_multiple(names),
                strict=True,
            ),
        )

    async def get(self, name: rn.RegisterName) -> "Result[Any]":
        """Get named register from device."""
        return (await self.get_multiple([name]))[0]

    async def set(
        self,
        name: rn.RegisterName,
        value: Any,  # noqa: ANN401
    ) -> bool:
        """Set named register on device."""
        try:
            reg = REGISTERS[name]
        except KeyError as err:
            msg = "Invalid Register Name"
            raise ValueError(msg) from err

        if not reg.writeable:
            msg = "Register is not writable"
            raise WriteException(msg)

        return await self._write_registers(reg, reg.encode(value))

    def _validate_data_to_write(self, register: "RegisterDefinition[Any]", values: tuple[Any, ...]) -> None:
        """Validate if the data to write is valid."""
        encoded_value_to_write = struct.pack(f">{register.format}", *values)
        if len(encoded_value_to_write) != register.length * 2:  # 2 bytes per register
            msg = "Wrong number of registers to write"
            raise WriteException(msg)

    async def _write_registers(
        self,
        register: "RegisterDefinition[Any]",
        values: tuple[Any, ...],
    ) -> bool:
        """Async write register to device."""
        self._validate_data_to_write(register, values)
        try:
            if register.length == 1:
                _LOGGER.debug(
                    "Writing to %d: single value '%s' on server %d",
                    register.register,
                    values[0],
                    self.unit_id,
                )

                response = await self.write_single_register(register.register, values[0])

                success: bool = response == values[0]
            else:
                _LOGGER.debug(
                    "Writing to %d: values '%s' on server %d",
                    register.register,
                    values,
                    self.unit_id,
                )

                registers_written = await self.write_struct_format(
                    register.register,
                    values,
                    format_struct=f">{register.format}",
                )

                success = registers_written == register.length

        except PermissionDeniedError:
            raise
        except IllegalDataAddressError as e:
            msg = (
                f"Failed to write value {values} to register {register} due to IllegalDataAddress. "
                "Assuming permission problem."
            )
            raise PermissionDeniedError(PermissionDeniedError.error_code, e.function_code) from e
        except ModbusResponseError as e:
            msg = f"Failed to write value {values} to register {register}: {e.error_code:02x}"
            raise WriteException(msg, modbus_exception_code=e.error_code) from e
        except ModbusConnectionError as err:
            _LOGGER.exception("Connection error while writing to register %s", register)
            raise ConnectionInterruptedException(err) from err
        return success



"""Base for classes that represent a single Huawei Solar device."""

import asyncio
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Self

from huawei_solar import register_names as rn
from huawei_solar.const import MAX_BATCHED_REGISTERS_COUNT, MAX_BATCHED_REGISTERS_GAP
from huawei_solar.exceptions import (
    HuaweiSolarException,
    InvalidCredentials,
)
from huawei_solar.modbus_pdu import PermissionDeniedError
from huawei_solar.registers import REGISTERS

if TYPE_CHECKING:
    from huawei_solar.modbus_client import AsyncHuaweiSolarClient
    from huawei_solar.register_definitions import Result

_LOGGER = logging.getLogger(__name__)

HEARTBEAT_INTERVAL = 15


WRITE_TEST_REGISTER = rn.TIME_ZONE


class HuaweiSolarDevice(ABC):
    """A higher-level interface making it easier to interact with a Huawei Solar inverter."""

    model_name: str
    serial_number: str
    update_lock: asyncio.Lock
    primary_device: "HuaweiSolarDevice | None" = None

    def __init__(
        self,
        client: "AsyncHuaweiSolarClient",
        model_name: str,
        *,
        primary_device: "HuaweiSolarDevice | None" = None,
    ) -> None:
        """DO NOT USE THIS CONSTRUCTOR DIRECTLY. Use create() method instead."""
        self.client = client
        self.model_name = model_name
        self.update_lock = primary_device.update_lock if primary_device else asyncio.Lock()
        self.primary_device = primary_device

    @classmethod
    async def create(
        cls,
        client: "AsyncHuaweiSolarClient",
        *,
        model_name: str,
        primary_device: "HuaweiSolarDevice | None" = None,
    ) -> Self:
        """Create instance with the necessary information."""
        device = cls(
            client,
            model_name,
            primary_device=primary_device,
        )

        await device._populate_additional_fields()

        return device

    @abstractmethod
    async def _populate_additional_fields(self) -> None:
        """Allow subclass to populate additional fields with information."""

    @classmethod
    @abstractmethod
    def supports_device(cls, model_name: str) -> bool:
        """Check if this class support the given device."""

    def _handle_batch_read_error(
        self,
        _queried_register_names: list[rn.RegisterName],
        exc: HuaweiSolarException,
    ) -> None:
        """Handle read errors in get."""
        raise exc

    def _detect_state_changes(self, new_values: "dict[rn.RegisterName, Result[Any]]") -> None:  # noqa: B027
        """Update state based on result of batch_update query.

        Used by subclasses to detect important changes.
        """

    async def _filter_registers(self, register_names: list[rn.RegisterName]) -> list[rn.RegisterName]:
        """Filter registers being requested in batch_update.

        Used by subclasses to prevent read-errors in certain cases.
        """
        return register_names

    def _transform_register_values(
        self,
        register_name: rn.RegisterName,  # noqa: ARG002
        result: "Result[Any]",
    ) -> "Result[Any]":
        """Optionally Transform the value of a register before returning it."""
        return result

    async def batch_update(self, register_names: list[rn.RegisterName]) -> "dict[rn.RegisterName, Result[Any]]":
        """Efficiently retrieve the values of all the registers passed in register_names.

        This method adds intelligence on top of read_multiple to only batch together
        registers that are close together in the inverter's memory map.
        """
        if unknown_registers := {register_name for register_name in register_names if register_name not in REGISTERS}:
            _LOGGER.warning(
                "Unknown register name passed to batch_update: %s",
                ", ".join(str(rn) for rn in unknown_registers),
            )

        class _Register:
            name: rn.RegisterName
            register_start: int
            register_end: int

            def __init__(self, regname: rn.RegisterName) -> None:
                self.name = regname

                reg = REGISTERS[regname]
                self.register_start = reg.register
                self.register_end = reg.register + reg.length - 1

        registers = [_Register(rn) for rn in register_names]

        registers.sort(key=lambda rd: rd.register_start)

        async with self.update_lock:
            result = {}
            first_idx = 0
            last_idx = 0

            while first_idx < len(registers):
                # Batch together registers:
                # - as long as the total amount of registers doesn't exceed 64
                # - as long as the gap between registers is not more than 16

                while (
                    last_idx + 1 < len(registers)
                    and registers[last_idx + 1].register_end - registers[first_idx].register_start
                    <= MAX_BATCHED_REGISTERS_COUNT
                    and registers[last_idx + 1].register_start - registers[last_idx].register_end
                    < MAX_BATCHED_REGISTERS_GAP
                ):
                    last_idx += 1

                register_names_to_query = [reg.name for reg in registers[first_idx : last_idx + 1]]
                register_names_to_query = await self._filter_registers(
                    register_names_to_query,
                )
                _LOGGER.debug(
                    "Batch update of the following registers: %s",
                    ", ".join(register_names_to_query),
                )

                try:
                    values = await self.client.get_multiple_as_dict(register_names_to_query)
                except HuaweiSolarException as exc:
                    self._handle_batch_read_error(register_names_to_query, exc)
                    values = {}

                self._detect_state_changes(values)
                result.update(values)

                first_idx = last_idx + 1
                last_idx = first_idx

            for key, value in result.items():
                result[key] = self._transform_register_values(key, value)

            return result

    async def stop(self) -> bool:
        """Stop the device connection."""
        if not self.primary_device:
            # we are the primary device, so we should also stop the client
            await self.client.disconnect()

        return True

    async def get(self, name: rn.RegisterName) -> "Result[Any]":
        """Get the value of a certain register."""
        return await self.client.get(name)

    async def set(self, name: rn.RegisterName, value: Any) -> bool:  # noqa: ANN401
        """Set a register to a certain value."""
        return await self.client.set(name, value)


class HuaweiSolarDeviceWithLogin(HuaweiSolarDevice, ABC):
    """A HuaweiSolarDevice that requires login to read any registers."""

    def __init__(
        self,
        client: "AsyncHuaweiSolarClient",
        model_name: str,
        *,
        primary_device: "HuaweiSolarDevice | None" = None,
    ) -> None:
        """Initialize with per-instance lock and login state."""
        super().__init__(client, model_name, primary_device=primary_device)
        self.__login_lock = asyncio.Lock()
        self.__heartbeat_enabled = False
        self.__heartbeat_task: asyncio.Task[None] | None = None
        self.__username: str | None = None
        self.__password: str | None = None

    async def ensure_logged_in(self, *, force: bool = False) -> bool:
        """Check if it is necessary to login and performs the necessary login sequence if needed."""
        async with self.__login_lock:
            if force:
                _LOGGER.debug(
                    "Forcefully stopping any heartbeat task (if any is still running)",
                )
                self.stop_heartbeat()

            if self.__username and not self.__heartbeat_enabled:
                _LOGGER.debug(
                    "Currently not logged in: logging in now and starting heartbeat",
                )
                if not self.__password:
                    msg = "Password must be set before logging in"
                    raise InvalidCredentials(msg)
                if not await self.client.login(self.__username, self.__password):
                    raise InvalidCredentials

                self.start_heartbeat()

        return True

    async def login(self, username: str, password: str) -> bool:
        """Perform the login-sequence with the provided username/password."""
        async with self.__login_lock:
            if not await self.client.login(username, password):
                raise InvalidCredentials

            # save the correct login credentials
            self.__username = username
            self.__password = password
            self.start_heartbeat()

        return True

    def stop_heartbeat(self) -> None:
        """Stop the running heartbeat task (if any)."""
        self.__heartbeat_enabled = False

        if self.__heartbeat_task:
            self.__heartbeat_task.cancel()

    def start_heartbeat(self) -> None:
        """Start the heartbeat thread to stay logged in."""
        if not self.__login_lock.locked():
            msg = "start_heartbeat should only be called from within the login_lock"
            raise RuntimeError(msg)

        if self.__heartbeat_task:
            self.stop_heartbeat()

        async def heartbeat() -> None:
            while self.__heartbeat_enabled:
                try:
                    self.__heartbeat_enabled = await self.client.heartbeat()
                    await asyncio.sleep(HEARTBEAT_INTERVAL)
                except HuaweiSolarException as err:
                    _LOGGER.warning("Heartbeat stopped because of, %s", err)
                    self.__heartbeat_enabled = False

        self.__heartbeat_enabled = True
        self.__heartbeat_task = asyncio.create_task(heartbeat())

    async def stop(self) -> bool:
        """Stop the device."""
        self.stop_heartbeat()

        return await super().stop()

    async def read_file(self, file_type: int, customized_data: bytes | None = None) -> bytes:
        """Wrap `get_file` from `AsyncHuaweiSolarClient`.

        This method adds a retry-logic for when the login-sequence needs to be repeated.
        """
        logged_in = await self.ensure_logged_in()

        if not logged_in:
            _LOGGER.warning(
                "Could not login, reading file %x will probably fail",
                file_type,
            )

        try:
            async with self.update_lock:
                return await self.client.get_file(file_type, customized_data)
        except PermissionDeniedError:
            if self.__username:
                logged_in = await self.ensure_logged_in(force=True)

                if not logged_in:
                    _LOGGER.exception("Could not login to read file %x", file_type)
                    raise

                async with self.update_lock:
                    return await self.client.get_file(
                        file_type,
                        customized_data,
                    )

            # we have no login-credentials available, pass on permission error
            raise

    ############################
    # Everything write-related #
    ############################

    async def has_write_permission(self) -> bool:
        """Test write permission by getting the time zone and trying to write that same value back to the inverter."""
        try:
            result = await self.client.get(WRITE_TEST_REGISTER)

            await super().set(WRITE_TEST_REGISTER, result.value)
        except PermissionDeniedError:
            return False
        else:
            return True

    async def set(self, name: rn.RegisterName, value: Any) -> bool:  # noqa: ANN401
        """Set a register to a certain value."""
        logged_in = await self.ensure_logged_in()  # we must login again before trying to set the value

        if not logged_in:
            _LOGGER.warning("Could not login, setting, %s will probably fail", name)

        if self.__heartbeat_enabled:
            try:
                await self.client.heartbeat()
            except HuaweiSolarException:
                _LOGGER.warning("Failed to perform heartbeat before write")

        try:
            return await super().set(name, value)
        except PermissionDeniedError:
            if self.__username:
                logged_in = await self.ensure_logged_in(force=True)

                if not logged_in:
                    _LOGGER.exception("Could not login to set %s", name)
                    raise

                # Force a heartbeat first when connected with username/password credentials
                await self.client.heartbeat()

                return await super().set(name, value)

            # we have no login-credentials available, pass on permission error
            raise



"""Test file for SUN2000Device."""

import asyncio
import logging

from huawei_solar import (
    SUN2000Device,
    create_device_instance,
    create_tcp_client,
    get_device_identifiers,
    get_device_infos,
)
from huawei_solar import register_names as rn

loop = asyncio.new_event_loop()

logging.basicConfig(level=logging.DEBUG)


async def test() -> None:
    """Run test."""
    client = create_tcp_client(host="192.168.1.1", port=503)

    print(await get_device_identifiers(client))
    print(await get_device_infos(client))

    device = await create_device_instance(client)

    assert isinstance(device, SUN2000Device)
    await device.login("installer", "00000a")
    print(
        await device.batch_update(
            [
                rn.ACTIVE_POWER_FIXED_VALUE_DERATING,
                rn.ACTIVE_POWER_PERCENTAGE_DERATING,
                rn.STORAGE_CAPACITY_CONTROL_MODE,
                rn.STORAGE_CAPACITY_CONTROL_SOC_PEAK_SHAVING,
                rn.STORAGE_CAPACITY_CONTROL_PERIODS,
            ],
        ),
    )
    print(await device.get(rn.ACTIVE_POWER_FIXED_VALUE_DERATING))
    print(await device.get(rn.ACTIVE_POWER_PERCENTAGE_DERATING))

    print(await device.get(rn.STORAGE_CAPACITY_CONTROL_MODE))
    print(await device.get(rn.STORAGE_CAPACITY_CONTROL_SOC_PEAK_SHAVING))
    print(await device.get(rn.STORAGE_CAPACITY_CONTROL_PERIODS))

    print(await device.get_optimizer_system_information_data())
    print(await device.get_latest_optimizer_history_data())

    await device.stop()


loop.run_until_complete(test())



"""Tests for the AsyncHuaweiSolarClient class."""

import struct
from unittest.mock import patch

import huawei_solar.register_names as rn
import huawei_solar.register_values as rv
import pytest
from huawei_solar.exceptions import DecodeError
from huawei_solar.modbus_client import AsyncHuaweiSolarClient
from huawei_solar.register_values import GridCode


async def test_get_model_name(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.MODEL_NAME)
    assert result.value == "SUN2000-3KTL-L1"
    assert result.unit is None


async def test_get_invalid_model_name(huawei_solar: AsyncHuaweiSolarClient) -> None:
    # invalid utf-8 sequence from here:
    # https://stackoverflow.com/questions/1301402/example-invalid-utf8-string

    value = struct.pack(
        ">15H",
        *[21333, 20018, int.from_bytes(b"\xa0\xa1"), 12336, 12333, 13131, 21580, 226, 0, 0, 0, 0, 0, 226, 10370],
    )

    with patch.object(huawei_solar, "execute", return_value=value):
        result = await huawei_solar.get(rn.MODEL_NAME)
        # Invalid UTF-8 bytes are replaced with backslash escapes
        assert "\\xa0\\xa1" in result.value


async def test_get_serial_number(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.SERIAL_NUMBER)
    assert result.value == "HV3021621085"
    assert result.unit is None


async def test_get_multiple(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get_multiple([rn.MODEL_NAME, rn.SERIAL_NUMBER])
    assert result[0].value == "SUN2000-3KTL-L1"
    assert result[0].unit is None

    assert result[1].value == "HV3021621085"
    assert result[1].unit is None


async def test_get_model_id(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.MODEL_ID)
    assert result.value == 348
    assert result.unit is None


async def test_get_nb_pv_strings(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.NB_PV_STRINGS)
    assert result.value == 2
    assert result.unit is None


async def test_get_nb_mpp_tracks(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.NB_MPP_TRACKS)
    assert result.value == 2
    assert result.unit is None


async def test_get_rated_power(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.RATED_POWER)
    assert result.value == 3000
    assert result.unit == "W"


async def test_get_p_max(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.P_MAX)
    assert result.value == 3300
    assert result.unit == "W"


async def test_get_s_max(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.S_MAX)
    assert result.value == 3300
    assert result.unit == "VA"


async def test_get_q_max_out(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.Q_MAX_OUT)
    assert result.value == 1980
    assert result.unit == "var"


async def test_get_q_max_in(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.Q_MAX_IN)
    assert result.value == -1980
    assert result.unit == "var"


async def test_get_state_1(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.STATE_1)
    assert result.value == ["Standby"]
    assert result.unit is None


async def test_get_state_1_extra_bits_set(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x7c\x00",
    ):
        result = await huawei_solar.get(rn.STATE_1)
        assert result.value == []
        assert result.unit is None


async def test_get_state_2(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.STATE_2)
    assert result.value == ["Locked", "PV disconnected", "No DSP data collection"]
    assert result.unit is None


async def test_get_state_2_extra_bits_set(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x7f\xf8",
    ):
        result = await huawei_solar.get(rn.STATE_2)

        assert result.unit is None


async def test_get_state_3(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.STATE_3)
    assert result.value == ["On-grid", "Off-grid switch disabled"]
    assert result.unit is None


async def test_get_state_3_extra_bits_set(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x7f\xf8\x7f\xff",
    ):
        result = await huawei_solar.get(rn.STATE_3)
        assert result.value == ["Off-grid", "Off-grid switch enabled"]
        assert result.unit is None


async def test_get_alarm_1_some(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.ALARM_1)
    expected_result = [
        rv.ALARM_CODES_1[1],
        rv.ALARM_CODES_1[256],
    ]
    assert result.value == expected_result
    assert result.unit is None


async def test_get_alarm_1_none(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x00\x00",
    ):
        result = await huawei_solar.get(rn.ALARM_1)
        assert result.value == []
        assert result.unit is None


async def test_get_alarm_1_all(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\xff\xff",
    ):
        result = await huawei_solar.get(rn.ALARM_1)
        expected_result = list(rv.ALARM_CODES_1.values())
        assert result.value == expected_result
        assert result.unit is None


async def test_get_alarm_2_some(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.ALARM_2)
    expected_result = [
        rv.ALARM_CODES_2[2],
        rv.ALARM_CODES_2[512],
    ]
    assert result.value == expected_result
    assert result.unit is None


async def test_get_alarm_2_none(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x00\x00",
    ):
        result = await huawei_solar.get(rn.ALARM_2)
        expected_result: list[str] = []
        assert result.value == expected_result
        assert result.unit is None


async def test_get_alarm_2_all(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\xff\xff",
    ):
        result = await huawei_solar.get(rn.ALARM_2)
        expected_result = list(rv.ALARM_CODES_2.values())
        assert result.value == expected_result
        assert result.unit is None


async def test_get_alarm_3_some(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.ALARM_3)
    expected_result = list(rv.ALARM_CODES_3.values())[0:2] + list(rv.ALARM_CODES_3.values())[3:5]
    assert result.value == expected_result
    assert result.unit is None


async def test_get_alarm_3_almost_all(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x7f\xff",
    ):
        result = await huawei_solar.get(rn.ALARM_3)
        expected_result = list(rv.ALARM_CODES_3.values())[:-1]
        assert result.value == expected_result
        assert result.unit is None


async def test_get_alarm_3_3rd_octet_bits_set(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with patch.object(
        huawei_solar,
        "execute",
        return_value=b"\x0e\x00",
    ):
        result = await huawei_solar.get(rn.ALARM_3)
        expected_result = list(rv.ALARM_CODES_3.values())[9:12]
        assert result.value == expected_result
        assert result.unit is None


async def test_get_pv_01_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_01_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_pv_01_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_01_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_pv_02_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_02_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_pv_02_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_02_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_pv_03_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_03_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_pv_03_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_03_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_pv_04_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_04_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_pv_04_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_04_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PV_01_VOLTAGE)
    assert result.value == 0.0
    assert result.unit == "V"


async def test_get_input_power(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.INPUT_POWER)
    assert result.value == 0
    assert result.unit == "W"


async def test_get_grid_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.GRID_VOLTAGE)
    assert result.value == 0.0
    assert result.unit == "V"


async def test_get_line_voltage_a_b(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.LINE_VOLTAGE_A_B)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_line_voltage_b_c(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.LINE_VOLTAGE_B_C)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_line_voltage_c_a(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.LINE_VOLTAGE_C_A)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_line_phase_a_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_A_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_line_phase_b_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_B_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_line_phase_c_voltage(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_C_VOLTAGE)
    assert result.value == 0
    assert result.unit == "V"


async def test_get_grid_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.GRID_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_phase_a_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_A_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_phase_b_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_B_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_phase_c_current(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.PHASE_C_CURRENT)
    assert result.value == 0
    assert result.unit == "A"


async def test_get_day_active_power_peak(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.DAY_ACTIVE_POWER_PEAK)
    assert result.value == 225
    assert result.unit == "W"


async def test_get_active_power(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.ACTIVE_POWER)
    assert result.value == 0
    assert result.unit == "W"


async def test_get_reactive_power(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.REACTIVE_POWER)
    assert result.value == 0
    assert result.unit == "var"


async def test_get_power_factor(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.POWER_FACTOR)
    assert result.value == 0.0
    assert result.unit is None


async def test_get_grid_frequency(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.GRID_FREQUENCY)
    assert result.value == 0.0
    assert result.unit == "Hz"


async def test_get_efficiency(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.EFFICIENCY)
    assert result.value == 0.0
    assert result.unit == "%"


async def test_get_internal_temperature(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.INTERNAL_TEMPERATURE)
    assert result.value == 0.0
    assert result.unit == "°C"


async def test_get_insulation_resistance(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.INSULATION_RESISTANCE)
    assert result.value == 3.0
    assert result.unit == "MOhm"


async def test_get_device_status(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.DEVICE_STATUS)
    assert result.value == "Standby: no irradiation"
    assert result.unit is None


async def test_get_device_status_invalid(huawei_solar: AsyncHuaweiSolarClient) -> None:
    with (
        patch.object(
            huawei_solar,
            "execute",
            return_value=b"\x02\xff",
        ),
        pytest.raises(DecodeError),
    ):
        await huawei_solar.get(rn.DEVICE_STATUS)


async def test_get_fault_code(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.FAULT_CODE)
    assert result.value == 0
    assert result.unit is None


async def test_get_accumulated_yield_energy(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.ACCUMULATED_YIELD_ENERGY)
    assert result.value == 207.34
    assert result.unit == "kWh"


async def test_get_daily_yield_energy(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.DAILY_YIELD_ENERGY)
    assert result.value == 0.65
    assert result.unit == "kWh"


async def test_get_nb_optimizers(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.NB_OPTIMIZERS)
    assert result.value == 10
    assert result.unit is None


async def test_get_nb_online_optimizers(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.NB_ONLINE_OPTIMIZERS)
    assert result.value == 0
    assert result.unit is None


async def test_get_grid_code(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.GRID_CODE)
    expected_result = GridCode(standard="C10/11", country="Belgium")
    assert result.value == expected_result
    assert result.unit is None


async def test_get_time_zone(huawei_solar: AsyncHuaweiSolarClient) -> None:
    result = await huawei_solar.get(rn.TIME_ZONE)
    assert result.value == 60
    assert result.unit == "min"



"""Huawei SDongle device support."""

from huawei_solar import register_names as rn

from .base import HuaweiSolarDevice


class SDongleDevice(HuaweiSolarDevice):
    """An SDongle device."""

    @classmethod
    def supports_device(cls, model_name: str) -> bool:
        """Check if this class support the given device."""
        return model_name.startswith("SDongle")

    async def _populate_additional_fields(self) -> None:
        (
            model_name_result,
            serial_number_result,
        ) = await self.client.get_multiple(
            [
                rn.MODEL_NAME,
                rn.SERIAL_NUMBER,
            ],
        )
        self.model_name = model_name_result.value
        self.serial_number = serial_number_result.value



"""Check if the register config is correct."""

import logging

import pytest
from huawei_solar.registers import REGISTERS

from huawei_solar import register_names as rn
from huawei_solar.register_definitions import TargetDevice

_LOGGER = logging.getLogger(__name__)


@pytest.mark.parametrize("target_device", list(TargetDevice))
def test_register_config(target_device: TargetDevice) -> None:
    """Parse all REGISTERS and check for correct order and potential overlaps."""
    registers = [r for r in REGISTERS.values() if target_device in r.target_device]
    registers.sort(key=lambda x: x.register)

    for idx in range(1, len(registers)):
        if registers[idx].register in [32066, 32072, 40000]:
            # skip these registers, as they have multiple entries
            continue
        if registers[idx - 1].register + registers[idx - 1].length > registers[idx].register:
            msg = (
                f"Requested registers must be in monotonically increasing order, "
                f"but {registers[idx - 1].register} + {registers[idx - 1].length} > {registers[idx].register}!"
            )
            raise ValueError(msg)
        if registers[idx - 1].register + registers[idx - 1].length < registers[idx].register:
            _LOGGER.info(
                "There is a gap between %s and %s!",
                {registers[idx - 1].register},
                {registers[idx].register},
            )


def test_all_register_names_have_a_register() -> None:
    """Check that all register names have a corresponding register defined."""
    register_names_without_register = []

    for register_name in dir(rn):
        if register_name.startswith("_"):
            continue
        register_value = getattr(rn, register_name)
        if not isinstance(register_value, str):
            continue
        if register_value not in REGISTERS:
            register_names_without_register.append(register_name)

    assert not register_names_without_register, f"Register names without a register: {register_names_without_register}"
